from mllm.structured_chat import SChat


def test_basic_schat():
    chat = SChat()
    chat.add_block("background", "You generating a random number.")
    chat.add_output_key("number", "int", "The generated number.")
    res = chat.complete()
    print(res)


from mllm import caching
from mllm.chat import Chat
from mllm.utils import p_map

def test_cached_chat():
    def get_random_response():
        chat = Chat()
        chat += "Introduce your self in one sentence."
        res = chat.complete(cache=True, options={"temperature": 1})
        return res
    num_1 = get_random_response()
    num_2 = get_random_response()
    assert num_1 == num_2

## See all test_image_chat_cache in test_image.py

def test_multi_thread_cache():
    print(caching._cache_kv.cache_path)
    def get_the_same_number(num):
        chat = Chat(f"Output the number {num}")
        res = chat.complete(cache=True)
        return res

    res = p_map(get_the_same_number, range(10))
    caching.close()


from mllm.cache.cache_service import caching

def display_chats(disable=False):
    from mllm.chat_logger import ChatLogger
    return ChatLogger(disable)

def log_chats(disable=False, show_table=True, save_path=None):
    from mllm.chat_logger import ChatLogger
    return ChatLogger(disable, save_path=save_path, show_table=show_table)

def display_embed_search(disable=False):
    from mllm.embedding.vector_store import EmbedSearchLogger
    return EmbedSearchLogger(disable)


def refresh_cache(disable=False):
    return caching.refresh_cache(disable)


import ast
import json
from dataclasses import dataclass

"""
# Parse
"""


@dataclass
class ParseOptions:
    cheap_model: str = "gpt-4o-mini"
    correct_json_by_model: bool = False


parse_options = ParseOptions()


class Parse:

    @staticmethod
    def obj(src: str):
        try:
            res = ast.literal_eval(src)
        except:
            raise ValueError(f"Invalid Python code: {src}")
        return res

    @staticmethod
    def dict(src: str):
        # find first {
        start = src.find("{")
        # find last }
        end = src.rfind("}")
        if start == -1 or end == -1:
            raise ValueError(f"Invalid json: {src}")
        json_src = src[start:end + 1]

        try:
            res = json.loads(json_src)
            return res
        except:
            pass

        try:
            res = ast.literal_eval(json_src)
            return res
        except:
            pass

        json_src_replaced = json_src.replace("\n", " ")
        try:
            res = json.loads(json_src_replaced)
            return res
        except:
            pass

        if parse_options.correct_json_by_model:
            try:
                res = parse_json_by_cheap_model(json_src)
                return res
            except:
                pass

        raise ValueError(f"Invalid json: {src}")

    @staticmethod
    def list(src):
        # find first [
        start = src.find("[")
        # find last ]
        end = src.rfind("]")
        if start == -1 or end == -1:
            raise ValueError(f"Invalid json: {src}")
        try:
            res = json.loads(src[start:end + 1])
        except:
            raise ValueError(f"Invalid json: {src}")
        return res

    @staticmethod
    def quotes(src: str):
        """
        Parse a string that is enclosed by quotes ```
        :param src:
        :return: contents of the quoted string
        """
        split_lines = src.split("\n")
        for start, line in enumerate(split_lines):
            if line.startswith("```"):
                title = line[3:]
                break
        else:
            raise ValueError("No closing quotes")
        for end in range(start + 1, len(split_lines)):
            if split_lines[end].startswith("```"):
                break
        else:
            raise ValueError("No closing quotes")
        contents = "\n".join(split_lines[start + 1: end])
        return contents

    @staticmethod
    def colon(src: str):
        """
        Parse a string that is separated by colons
        :param src:
        :return: the contents after the colon
        """
        split = src.split(":")
        if len(split) != 2:
            raise ValueError("Invalid colon string")
        contents = split[1].strip()
        return contents


def parse_json_by_cheap_model(json_src):
    """
    Correct a JSON dict with semantic errors using a model that support JSON model
    This is especially for claude models because they do not support JSON model
    :param json_src:
    :return:
    """
    from mllm import Chat
    prompt = f"""
You are required to correct a JSON dict with semantic errors. 
<raw_json>
{json_src}
</raw_json>
You should directly output the corrected JSON dict with a minimal modification.
"""
    chat = Chat(prompt)
    res = chat.complete(parse="dict", model=parse_options.cheap_model)
    return res



from __future__ import annotations

import sqlite3
import threading


class ConnPool:
    def __init__(self, db_path: str):
        self.db_path = db_path
        self.conn_pool = {}
        self.threads = {}
        self.max_conn = 50
        self.lock_for_closing = threading.Lock()

    def get_conn(self):
        current_thread = threading.current_thread()
        thread_id = current_thread.ident
        conn = self.conn_pool.get(thread_id)
        if conn is not None:
            return conn
        # wait for the lock
        with self.lock_for_closing:
            # Add a new connection if the thread does not have one
            if thread_id not in self.conn_pool:
                self.conn_pool[thread_id] = sqlite3.connect(self.db_path, check_same_thread=False)
                self.threads[thread_id] = current_thread
            if len(self.conn_pool) >= self.max_conn:
                # remove the connections that is not running
                for tid in list(self.threads.keys()):
                    if not self.threads[tid].is_alive():
                        del self.conn_pool[tid]
                        del self.threads[tid]
            return self.conn_pool[thread_id]

    def close_conn(self):
        with self.lock_for_closing:
            thread_id = threading.get_ident()
            if thread_id in self.conn_pool:
                self.conn_pool[thread_id].close()
                del self.conn_pool[thread_id]
                del self.threads[thread_id]

    def close_all(self):
        with self.lock_for_closing:
            for conn in self.conn_pool.values():
                conn.close()
            self.conn_pool = {}
            self.threads = {}



from __future__ import annotations

import os
import atexit
import sys
from typing import List

from mllm.cache.cache_embedding import CacheTableEmbed
from mllm.cache.cache_kv import CacheTableKV

def get_main_path():
    return os.path.abspath(sys.argv[0])


def get_cache_path(main_path: str, postfix):
    dir_name = main_path
    assert len(postfix) > 0
    postfix_str = ".".join(postfix)
    return os.path.join(dir_name, ".llm_cache", postfix_str + ".db")


class CacheService:

    def __init__(self):
        main_path = get_main_path()
        self.postfix_stack = [os.path.basename(main_path)]
        self.base_path = os.path.dirname(main_path)
        cache_path = get_cache_path(self.base_path, self.postfix_stack)
        self._cache_kv: CacheTableKV = CacheTableKV(cache_path)
        self.cache_embed: CacheTableEmbed = CacheTableEmbed(os.path.dirname(cache_path))
        self.cache_kv_other = {self._cache_kv.cache_path: self._cache_kv}
        self.cache_embed_other = {self._cache_kv.cache_path: self.cache_embed}
        self.cache_kv_disabled = False

    def read_kv_cache(self, key: str, type: str):
        if self.cache_kv_disabled:
            return None
        return self._cache_kv.read_cache(key, type)
    
    def disable_cache_kv(self):
        self.cache_kv_disabled = True

    def enable_cache_kv(self):
        self.cache_kv_disabled = False

    def save(self):
        self._cache_kv.save_all_cache_to_file()
        for cache_kv in self.cache_kv_other.values():
            cache_kv.save_all_cache_to_file()
        self.cache_embed.save_pending_cache()
        for cache_embed in self.cache_embed_other.values():
            cache_embed.save_pending_cache()

    def save_used(self):
        self._cache_kv.save_all_cache_to_file(filter_unused_cache=True)
        for cache_kv in self.cache_kv_other.values():
            cache_kv.save_all_cache_to_file(filter_unused_cache=True)
        # TODO: save only used embedding cache
        self.cache_embed.save_pending_cache()

    def close(self):
        self.cache_embed.close()
        for cache_embed in self.cache_embed_other.values():
            cache_embed.close()
        self._cache_kv.close()
        for cache_kv in self.cache_kv_other.values():
            cache_kv.close()

    def at_exit(self):
        self.save()
        self.close()

    def refresh_cache(self, disable=False):
        """
        Usage: with caching.refresh_cache():
                your codes
        :return: a context manager that refreshes the cache
        """
        return RefreshCacheContext(self._cache_kv, disable=disable)

    def disable_cache(self, disable=False):
        """
        Usage: with caching.disable_cache():
                your codes
        :return:
        """
        return DisableCacheContext(self._cache_kv, disable=disable)

    def _load_cache_on_path(self, postfix: List[str]):
        self._cache_kv.apply_cache_update()
        kv_cache_path = get_cache_path(self.base_path, postfix)
        cache_dir = os.path.dirname(kv_cache_path)

        if kv_cache_path in self.cache_kv_other:
            self._cache_kv = self.cache_kv_other[kv_cache_path]
        else:
            self._cache_kv = CacheTableKV(kv_cache_path)
            self.cache_kv_other[kv_cache_path] = self._cache_kv

        embed_cache_postfix = "".join(["."+p for p in postfix[1:]])

        if cache_dir in self.cache_embed_other:
            self.cache_embed = self.cache_embed_other[(cache_dir, embed_cache_postfix)]
        else:
            self.cache_embed = CacheTableEmbed(cache_dir, embed_cache_postfix)
            self.cache_embed_other[(cache_dir, embed_cache_postfix)] = self.cache_embed

    def push_postfix(self, postfix):
        self.postfix_stack.append(postfix)
        self._load_cache_on_path(self.postfix_stack)

    def pop_postfix(self):
        assert len(self.postfix_stack) >= 2
        self.postfix_stack.pop()
        self._load_cache_on_path(self.postfix_stack)

    def set_root_name(self, name: str):
        self.postfix_stack[0] = name
        self._load_cache_on_path(self.postfix_stack)

    def set_root_path(self, dir_path: str):
        self.base_path = dir_path
        self._load_cache_on_path(self.postfix_stack)


    def cache_env(self, postfix):
        return PostfixContext(self, postfix)


class PostfixContext:
    def __init__(self, cache_service: CacheService, postfix: str):
        self.cache_service = cache_service
        self.postfix = postfix

    def __enter__(self):
        self.cache_service.push_postfix(self.postfix)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cache_service.pop_postfix()


class RefreshCacheContext:
    def __init__(self, cache_kv: CacheTableKV, cache_type: str = "", disable=False):
        """
        :param cache_type: The type of cache to refresh. If type is "", then all cache will be
        refreshed
        """
        self.cache_type = cache_type
        self.cache_kv = cache_kv
        self.already_refreshed = False
        self.disable = True if disable != False else False

    def __enter__(self):
        if self.disable:
            return
        if self.cache_type != "":
            if self.cache_type not in self.cache_kv.types_to_refresh:
                self.cache_kv.types_to_refresh.add(self.cache_type)
            else:
                self.already_refreshed = True
        else:
            if self.cache_kv.refresh_all:
                self.already_refreshed = True
            else:
                self.cache_kv.refresh_all = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return
        if self.cache_type != "":
            if not self.already_refreshed:
                self.cache_kv.types_to_refresh.remove(self.cache_type)
        else:
            if not self.already_refreshed:
                self.cache_kv.refresh_all = False
        self.cache_kv.apply_cache_update()


class DisableCacheContext:
    def __init__(self, cache_kv: CacheTableKV, disable=False):
        self.cache_kv: CacheTableKV = cache_kv
        self.disable = True if disable != False else False

    def __enter__(self):
        if self.disable:
            return
        self.cache_kv.inactive = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return
        self.cache_kv.inactive = False


caching = CacheService()

# save cache on exit
atexit.register(caching.at_exit)



from typing import Dict

from mllm import Chat
from mllm.chat import remove_all_indents


class JsonSpec:
    def __init__(self):
        self.keys = []

    def add_output_key(self, key: str, type: str, requirement: str = ""):
        requirement = remove_all_indents(requirement)
        requirement = requirement.strip()
        self.keys.append(
            {
                "key": key,
                "type": type,
                "requirement": requirement
            }
        )

    def in_prompt(self):
        keys = []
        for key_item in self.keys:
            key = key_item["key"]
            type = key_item["type"]
            requirement = key_item["requirement"]
            if len(requirement) > 0:
                keys.append(f'"{key}"({type}): {requirement}')
            else:
                keys.append(f'"{key}"({type})')
        keys = "\n".join(keys)
        prompt = f"<output>\nYou are required to output a JSON dict with the following keys:\n{keys}\n</output>"
        return prompt


class Block:
    def __init__(self, tagname: str, text: str):
        self.tagname = tagname
        self.text = text

    def in_prompt(self):
        return f"<{self.tagname}>\n{self.text}\n</{self.tagname}>"


class SChat:
    """
    The class for structured chat.
    """

    def __init__(self):
        self.blocks = []
        self.output_requirement = JsonSpec()

    def add_block(self, tagname: str, text: str, dedent=True):
        if dedent:
            text = remove_all_indents(text.strip())
        self.blocks.append(Block(tagname, text))

    def add_output_key(self, key: str, type: str, document: str):
        self.output_requirement.add_output_key(key, type, document)

    def in_prompt(self):
        res = []
        for block in self.blocks:
            res.append(block.in_prompt())
        res.append(self.output_requirement.in_prompt())
        res_str = "\n".join(res)
        return res_str

    def get_chat(self):
        chat = Chat()
        for block in self.blocks:
            chat += block.in_prompt()
        chat += self.output_requirement.in_prompt()
        return chat

    def complete(self, model=None, cache=False, expensive=False, options=None) -> Dict:
        res = self.get_chat().complete(model=model, cache=cache, expensive=expensive,
                                       parse="dict", retry=True, options=options)
        return res



import hashlib
import os

import numpy as np
import sqlite3
from datetime import date

from mllm.cache.conn_pool import ConnPool


def get_hash(text: str):
    return hashlib.md5(text.encode()).hexdigest()

class CacheTableEmbed:
    def __init__(self, cache_dir: str, post_fix=""):
        self.cache_dir = cache_dir
        self.post_fix = post_fix
        self.pending_cache = {}
        db_path = self.get_db_path()
        db_exist = os.path.exists(db_path)
        self.conn_pool = ConnPool(self.get_db_path())
        if not db_exist:
            self.create_cache_table()

    @property
    def db_conn(self):
        return self.conn_pool.get_conn()

    def create_cache_table(self):
        cursor = self.db_conn.cursor()
        cursor.execute('''CREATE TABLE embedding_cache
                         (hash TEXT PRIMARY KEY, 
                         model_name text, 
                         date text,
                         embedding blob)''')
        self.db_conn.commit()

    def add_cache(self, model_name: str, text: str, embedding: np.ndarray):
        hash_value = get_hash(text)
        self.pending_cache[(model_name, hash_value)] = embedding

    def read_cache(self, model_name: str, text: str) -> np.ndarray:
        text_hash = get_hash(text)
        if (model_name, text_hash) in self.pending_cache:
            return self.pending_cache[(model_name, text_hash)]
        cursor = self.db_conn.cursor()
        cursor.execute("SELECT embedding FROM embedding_cache WHERE hash = ? AND model_name = ?", (text_hash, model_name))
        res = cursor.fetchone()
        if res is None:
            return None
        return np.frombuffer(res[0], dtype=np.float32)

    def get_db_path(self):
        return os.path.join(self.cache_dir, f"embedding_cache{self.post_fix}.db")

    def save_pending_cache(self):
        for (model_name, hash_value), embedding in self.pending_cache.items():
            cursor = self.db_conn.cursor()
            cursor.execute("INSERT OR REPLACE INTO embedding_cache VALUES (?, ?, ?, ?)", (hash_value, model_name, str(date.today()), embedding.tobytes()))
        self.db_conn.commit()

    def clear_cache_table(self):
        self.pending_cache = {}
        self.db_conn.execute("DELETE FROM embedding_cache")
        self.db_conn.commit()

    def close(self):
        self.conn_pool.close_all()



from __future__ import annotations

import html
import json

from mllm.display.show_html import show_json_table
from mllm.utils.logger import Logger


def get_chat_in_html(message_to_api: list[dict], additional_res: dict):
    res = []
    for entry in message_to_api:
        if isinstance(entry["content"], str):
            entry["content"] = [{"type": "text", "text": entry["content"]}]
        content = []
        for item in entry["content"]:
            if item["type"] == "text":
                text = item["text"]
                text = html.escape(text)
                text = text.replace("\n", "<br/>")
                text = text.replace(" ", "&nbsp;")
                content.append(text)
            elif item["type"] == "image_url":
                content.append("<image src='{}' style='max-height: 200px;'/>".format(
                    item["image_url"]["url"]))
        content = "<br/>".join(content)
        res.append(f"------{entry['role']}------<br/> {content}")
        res.append("""
<style>
details {
  user-select: none;
}

details>summary span.icon {
  width: 24px;
  height: 24px;
  transition: all 0.3s;
  margin-left: auto;
}

details[open] summary span.icon {
  transform: rotate(180deg);
}

summary {
  display: flex;
  cursor: pointer;
  background-color: #f1f1f1;
}

summary::-webkit-details-marker {
  display: none;
}
</style>    
""")
    additional_res_str = json.dumps(additional_res, indent=4)
    additional_res_str = html.escape(additional_res_str)
    res.append(f"""
<details>
  <summary>Additional info</summary>
  <pre>{additional_res_str}</pre>
</details>
""")

    return "<br/>".join(res)


class ChatLogger(Logger):
    active_loggers = []
    def __init__(self, disable=False, save_path=None, show_table=True):
        super().__init__(disable)
        self.save_path = save_path
        self.show_table = show_table


    def display_log(self):
        contents = [get_chat_in_html(message_to_api, additional_res) for message_to_api, additional_res in self.log_list]
        filenames = [caller_name.split("/")[-1] for caller_name in self.caller_list]
        info_list = []
        for i in range(len(contents)):
            content = contents[i]
            info_list.append({
                "filename": filenames[i],
                "content": content
            })
        if self.show_table:
            show_json_table(info_list)
        if self.save_path is not None:
            info_list_for_save = []
            for i in range(len(self.log_list)):
                message_to_api, additional_res = self.log_list[i]
                info_list_for_save.append({
                    "filename": filenames[i],
                    "content": message_to_api,
                    "additional_res": additional_res
                })
            with open(self.save_path, "w") as f:
                f.write(json.dumps(info_list_for_save, indent=4))



from time import sleep
from stqdm import stqdm
from mllm.utils import p_map

wait_for_1_second = lambda x: sleep(1) or x**2
p_map(wait_for_1_second, [1, 2, 3], n_workers=1, title="test", pbar_impl=stqdm)


import os


def get_llava_response(messages, options, model_name):
    prompt, images = extract_prompt_and_image(messages)
    assert len(images) == 1, "Only one image supported in llava model"
    image = images[0]
    try:
        import replicate
    except ImportError:
        raise ImportError("Please install the `replicate` package to use the llava model")
    if os.environ["REPLICATE_API_TOKEN"] is None:
        if os.environ["REPLICATE_API_KEY"] is None:
            raise ValueError("Please set the REPLICATE_API_KEY environment variable")
        os.environ["REPLICATE_API_TOKEN"] = os.environ["REPLICATE_API_KEY"]
    model_name = model_name.split("/")[1:]
    model_name = "/".join(model_name)
    output = replicate.run(
        model_name,
        input={
            "image": image,
            "top_p": 1,
            "prompt": prompt,
            "history": [],
            "max_tokens": 2048,
            "temperature": options["temperature"],
        }
    )

    res = "".join(output)
    return res


def get_llama_response(messages, options, model_name):
    try:
        import replicate
    except ImportError:
        raise ImportError("Please install the `replicate` package to use the llama model")
    model_name = model_name.split("/")[1:]
    model_name = "/".join(model_name)
    prompt, images = extract_prompt_and_image(messages)
    assert len(images) == 0, "Image not supported in llama model"
    output = replicate.run(
        model_name,
        input={
            "top_p": 1,
            "prompt": prompt,
            "history": [],
            "max_tokens": 2048*2,
            "temperature": options["temperature"],
        }
    )
    res = "".join(output)
    return res


def extract_prompt_and_image(messages):

    prompt = []
    images = []
    for message in messages:
        if message["role"] == "user":
            content = message["content"]
            if isinstance(content, str):
                prompt.append(content)
            else:
                for item in content:
                    if item["type"] == "text":
                        prompt.append(item["text"])
                    elif item["type"] == "image_url":
                        image = item["image_url"]["url"]
                        images.append(image)
        else:
            if isinstance(message["content"], str):
                prompt.append(f"<{message['role']}>" + message['content'] + f"<{message['role']}>")
            elif isinstance(message["content"], dict):
                prompt.append(f"<{message['role']}>" + message["content"]["text"] + f"<{message['role']}>")
            elif isinstance(message["content"], list):
                for item in message["content"]:
                    if item["type"] == "text":
                        prompt.append(f"<{message['role']}>" + item["text"] + f"<{message['role']}>")
                    elif item["type"] == "image_url":
                        raise ValueError("Image URL not supported in assistant or system prompts.")

    prompt = "\n".join(prompt)
    return prompt, images


special_models = {
    "replicate/yorickvp/llava-v1.6-34b:41ecfbfb261e6c1adf3ad896c9066ca98346996d7c4045c5bc944a79d430f174": get_llava_response
}


def get_special_model_handler(model_name):
    if not model_name.startswith("replicate/"):
        return None
    if "yorickvp/llava" in model_name:
        return lambda messages, options: get_llava_response(messages, options, model_name)
    if "llama" in model_name:
        return lambda messages, options: get_llama_response(messages, options, model_name)




from mllm import caching

def test_env():
    with caching.refresh_cache():
        with caching.cache_env("t"):
            cache = caching.read_kv_cache("123", "int")
            cache.set_cache("1234")

        cache = caching.read_kv_cache("123", "int")
        cache.set_cache("12345")

        with caching.cache_env("t"):
            cache = caching.read_kv_cache("123", "int")
            assert cache.value == "1234"


from __future__ import annotations

import hashlib
import json
import os
from datetime import date
from typing import Dict, Optional, Set

from mllm.cache.conn_pool import ConnPool


def get_hash(data: any) -> str:
    return hashlib.sha1(json.dumps(data).encode("utf-8")).hexdigest()


class Cache:
    """
    The class for storing cache.
    """

    def __init__(self, value, hash: str, input: any, type: str,
                 meta: Optional[Dict] = None):
        self.value = value
        # self.hash should be the same as get_hash(input, type)
        self.hash: str = hash
        self.input: any = ""#input
        self.type: str = type
        self.meta = meta or {}

    def get_self_dict(self):
        return {
            "value": self.value,
            "hash": self.hash,
            "input": self.input,
            "type": self.type,
            "meta": self.meta
        }

    def set_cache(self, value: any, meta: Optional[Dict] = None):
        self.value = value
        self.meta = meta

    def is_valid(self):
        return self.value is not None


CacheTable = Dict[str, Cache]


class CacheTableKV:
    def __init__(self, cache_path: str):
        self.cache_path = cache_path
        # List of pending cache
        self.pending_cache: Dict[str, Cache] = {}
        # List of active cache. Used for garbage collection
        self.active_cache_hash: Set[str] = set()
        #
        self.types_to_refresh: Set[str] = set()
        #
        self.refresh_all: bool = False
        #
        self.inactive = False

        db_path = cache_path
        db_exist = os.path.exists(db_path)
        db_dir = os.path.dirname(db_path)
        # create the directory if not exist
        if len(db_dir) != 0:
            if not os.path.exists(db_dir):
                os.makedirs(db_dir)
        self.conn_pool = ConnPool(db_path)
        if not db_exist:
            self.create_cache_table()

    def create_cache_table(self):
        db_conn = self.conn_pool.get_conn()
        cursor = db_conn.cursor()
        cursor.execute('''CREATE TABLE cache_table
                         (hash TEXT PRIMARY KEY,
                         type text,
                         value text,
                         date text,
                         meta text)''')
        db_conn.commit()

    def save_all_cache_to_file(self, filter_unused_cache=False):
        if filter_unused_cache:
            n_remove = self.filter_unused_cache()
            if n_remove > 0:
                print(f"Removed {n_remove} unused cache")
        self.apply_cache_update()

    def filter_unused_cache(self) -> int:
        # remove all the rows whose hash is not in self.active_cache_hash
        db_conn = self.conn_pool.get_conn()
        cursor = db_conn.cursor()
        n_rows = cursor.execute("SELECT COUNT(*) FROM cache_table").fetchone()[0]
        cursor.execute("DELETE FROM cache_table WHERE hash NOT IN ({})".format(
            ",".join(["?"] * len(self.active_cache_hash))), tuple(self.active_cache_hash))
        n_rows_after = cursor.execute("SELECT COUNT(*) FROM cache_table").fetchone()[0]
        db_conn.commit()
        return n_rows - n_rows_after

    def read_cache(self, input: any, type: str, create_cache=True) -> Cache | None:

        if self.inactive:
            return None
        hash = get_hash(input)

        db_conn = self.conn_pool.get_conn()
        cursor = db_conn.cursor()
        cursor.execute("SELECT value, meta FROM cache_table WHERE hash = ? AND type = ?", (hash, type))
        db_conn.commit()
        res = cursor.fetchone()

        meta = {}
        cache_value = None
        if res is None:
            un_hit = True
        else:
            cache_value, meta = res
            if meta != "None":
                meta = json.loads(meta)
            else:
                meta = {}
            un_hit = False

        if type in self.types_to_refresh or self.refresh_all:
            un_hit = True
        if un_hit:
            if create_cache:
                new_cache = Cache(None, hash, input, type, meta)
                self.add_cache(new_cache)
                return new_cache
            else:
                return None

        cache_hit = Cache(cache_value, hash, input, type, meta)
        self.active_cache_hash.add(hash)

        return cache_hit

    def add_cache(self, cache: Cache):
        self.pending_cache[cache.hash] = cache

    def apply_cache_update(self):
        remaining_cache = {}
        db_conn = self.conn_pool.get_conn()
        cursor = db_conn.cursor()
        for hash, cache in list(self.pending_cache.items()):
            if cache.is_valid():
                self.active_cache_hash.add(cache.hash)
                cursor.execute("INSERT OR REPLACE INTO cache_table VALUES (?, ?, ?, ?, ?)",
                               (cache.hash, cache.type, cache.value, str(date.today()), str(cache.meta)))

            else:
                remaining_cache[hash] = cache
        db_conn.commit()
        self.pending_cache = remaining_cache

    def discard_cache_update(self):
        self.pending_cache = {}

    def close(self):
        self.conn_pool.close_all()







import threading
import time

from mllm.cache.conn_pool import ConnPool


def test():
    conn_pool = ConnPool("./test.db")
    conn_pool.max_conn = 5
    def get_conn():
        conn_pool.get_conn()
        time.sleep(5)

    for i in range(20):
        thread = threading.Thread(target=get_conn)
        thread.start()


from dataclasses import dataclass


@dataclass
class DefaultModels:
    normal: str = "gpt-4o-mini"
    expensive: str = "gpt-4o"
    vision: str = "gpt-4o"
    embedding: str = "text-embedding-3-large"

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


@dataclass
class DefaultOptions:
    temperature: float = 0.2
    max_tokens: int = None
    frequency_penalty: float = None
    timeout: int = 600
    seed: int = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


default_models = DefaultModels()

default_options = DefaultOptions()

if __name__ == '__main__':
    print(default_options.get_dict())



from __future__ import annotations

from typing import List, Dict

import numpy as np
from litellm import embedding

from mllm.cache.cache_embedding import CacheTableEmbed
from mllm.cache.cache_service import caching
from mllm.config import default_models


class LazyEmbedding:
    lazy_embeddings: Dict[str, List[LazyEmbedding]] = {}

    def __init__(self, src: str, model: str, cache_embed: CacheTableEmbed):
        self.src = src
        self.model = model
        self.cache_embed: CacheTableEmbed = cache_embed
        self.embedding = None
        if model not in LazyEmbedding.lazy_embeddings:
            LazyEmbedding.lazy_embeddings[model] = [self]
        else:
            LazyEmbedding.lazy_embeddings[model].append(self)


    def __array__(self, *array_args, **array_kwargs):
        if self.embedding is None:
            self.flush()
        if len(array_args) > 0:
            return np.array(self.embedding, *array_args, **array_kwargs)
        return self.embedding

    def to_list(self):
        return list(self.__array__())

    def flush(self):
        lazy_embeddings = LazyEmbedding.lazy_embeddings[self.model]
        embeddings = _get_embeddings(self.model, [le.src for le in lazy_embeddings])
        for i, lazy_embedding in enumerate(lazy_embeddings):
            lazy_embedding.embedding = embeddings[i]
            self.cache_embed.add_cache(self.model, lazy_embedding.src, lazy_embedding.embedding)
        LazyEmbedding.lazy_embeddings[self.model] = []

    def __str__(self):
        return str(self.__array__())

    def __repr__(self):
        return repr(self.__array__())

    def __mul__(self, other):
        return self.__array__() * other


def get_embeddings(texts: list[str], model=None, lazy=True) -> list[list[float]]:
    if model is None:
        model = default_models["embedding"]
    cache_embed = caching.cache_embed

    embeddings = []
    index_for_eval = []
    texts_without_cache = []
    for i, text in enumerate(texts):
        if len(text) == 0:
            raise ValueError("Text cannot be empty")
            embeddings.append(0)
            continue
        embedding_from_cache = cache_embed.read_cache(model, text)
        if embedding_from_cache is None:
            texts_without_cache.append(text)
            embeddings.append(None)
            index_for_eval.append(i)
        else:
            embeddings.append(embedding_from_cache)
    if len(texts_without_cache) > 0:
        if not lazy:
            res = _get_embeddings(model, texts_without_cache)
            for i, r in zip(index_for_eval, res):
                cache_embed.add_cache(model, texts[i], r)
                embeddings[i] = r
        else:
            for i in index_for_eval:
                le = LazyEmbedding(texts[i], model, cache_embed)
                embeddings[i] = le

    return embeddings


def _get_embeddings(model, texts_without_cache):
    if len(texts_without_cache) > 2000:
        res = []
        for i in range(0, len(texts_without_cache), 2000):
            raw_res = embedding(model, input=texts_without_cache[i:i + 2000])
            res += [np.array(r['embedding'], dtype=np.float32) for r in raw_res.data]
        return res
    else:
        raw_res = embedding(model, input=texts_without_cache)
        res = [np.array(r['embedding'], dtype=np.float32) for r in raw_res.data]
        return res




import sys


def debugger_is_active() -> bool:
    """Return if the debugger is currently active"""
    return hasattr(sys, 'gettrace') and sys.gettrace() is not None

from .maps import parallel_map, p_map